#!/usr/bin/env python3

import sys
from pprint import pprint
from collections import OrderedDict,Counter
import itertools as it
import datetime
from sequence import Sequence

SPLITSIZE=2

def basecmp(a,b):
    return a==b or a=='N' or b=='N'

class CollapseTree:
    def __init__(self,iterable=(),_depth=0):
        self.seqs = {}
        self.children = {}
        self.depth=_depth
        for seq in iterable:
            self.add(seq)

    def split(self):
        newseqs = {}
        for seq in self.seqs:
            if len(seq)==self.depth:
                newseqs[seq]=self.seqs[seq]
            else:
                if seq[self.depth] not in self.children:
                    self.children[seq[self.depth]] = CollapseTree(_depth=self.depth+1)
                self.children[seq[self.depth]].add(self.seqs[seq],seq)
        self.seqs=newseqs

    def add(self,rec,key=None):
        if key is None: key=rec.seq
        if key in self.seqs:
            self.seqs[key] = combine(self.seqs[key],rec)
        elif (self.children is not None and
              len(key)>self.depth and
              key[self.depth] in self.children):
            self.children[key[self.depth]].add(rec,key)
        else:
            self.seqs[key] = rec
        if len(self.seqs)>SPLITSIZE:
            self.split()

    def __iter__(self):
        yield from self.seqs.items()
        for d in self.children.values():
            yield from d

    def neighbors(self,seq):
        def nfil(x): return similar(seq,x)
        yield from filter(nfil,self.seqs)
        if len(seq)>self.depth:
            for d in [self.children[c]
                      for c in self.children
                      if basecmp(c,seq[self.depth])]:
                yield from d.neighbors(seq)

    def fixedseqs(self):
        ncd = CollapseTree()
        for seq,rec in self:
            newseq=seq
            if 'N' in seq:
                newseq=consensus(seq,self.neighbors(seq))
                rec.seq=newseq
            ncd.add(rec,newseq)
        return ncd


    def __len__(self):
        return len(self.seqs)+sum(len(c) for c in self.children.values())

def combine(f1,f2):
    targ = (f1 if f1.val >= f2.val else f2).copy()
    targ.conscount =  f1.conscount+f2.conscount
    targ.dupcount = f1.dupcount+f2.dupcount
    targ.key = f1.key+f2.key
    return targ

def similar(s1,s2):
    return len(s1)==len(s2) and all(basecmp(n1,n2)
                                    for n1,n2 in zip(s1,s2))

def basecon(bases):
    bases.discard('N')
    if len(bases)==1: return bases.pop()
    else: return 'N'

def consensus(seq,A):
    return ''.join((basecon(set(B)) if s=='N' else s) for (s,B) in zip(seq,zip(*A)))

if __name__=="__main__":
    pass
